(* +--------------------------------------------------------------+
   |                        Calcul multiprcision                 |
   |                                                              |
   +--------------------------------------------------------------+ *)

let BASE = 10;;
type nombre = {lg : int; ch : int list};; (* longueur, liste des chiffres *)

let zro = {lg = 0; ch = [] } (* constantes usuelles *)
and un   = {lg = 1; ch = [1]}
and deux = {lg = 1; ch = [2]};;


                        (* +---------------+
                           |  Utilitaires  |
                           +---------------+ *)

(* chiffre des units, reste du nombre *)
let dcompose(a) = if a = zro then 0,zro else hd(a.ch), {lg = a.lg - 1; ch = tl(a.ch)};;

(* x = chiffre, a = nombre multiprcision, retourne x + a*BASE *)
let assemble x a = if (a.lg > 0) or (x > 0) then {lg = a.lg + 1; ch = x::a.ch} else zro;;

(* dcalage de n chiffres  droite *)
let rec prefix >> a n = if n = 0 then a else assemble 0 (a >> (n-1));;

(* dcalage de n chiffres  gauche *)
let rec prefix << a n =
  if n = 0 then zro, a
  else let x,y = dcompose(a) in
       let u,v = y << (n-1)   in (assemble x u), v
;;

(* conversion int -> nombre *)
let rec nombre_of_int(x) =
  if x < 0 then failwith "nombre ngatif"
  else if x = 0 then zro
  else assemble (x mod BASE) (nombre_of_int(x/BASE))
;;

(* conversion nombre -> int (peut dborder) *)
let rec int_of_nombre(a) =
  if a = zro then 0
  else let x,y = dcompose(a) in x + BASE*(int_of_nombre y)
;;

(* nombre alatoire de n chiffres *)
let rec random_nombre n =
  if n = 0 then zro
  else if n = 1 then {lg = 1; ch = [1 + random__int(BASE-1)]}
  else assemble (random__int BASE) (random_nombre (n-1))
;;

(* affichage *)
let affiche(a) =
  let rec loop(l) = match l with u::v -> loop(v); print_int(u) |[] -> () in
  loop(a.ch); print_newline()
;;

(* rsultat d'une comparaison *)
type cmpres = Suprieur | Egal | Infrieur;;


                         (* +------------+
                            |  Addition  |
                            +------------+ *)

(* a,b = nombres multiprcision, c = retenue, retourne a + b + c *)
let rec add_c a b c =
  if (a = zro) & (b = zro) then nombre_of_int(c)
  else begin
    let a0,a1 = dcompose(a) in
    let b0,b1 = dcompose(b) in
    let x     = a0 + b0 + c  in
    if x < BASE then assemble x          (add_c a1 b1 0)
                else assemble (x - BASE) (add_c a1 b1 1)
  end
;;

(* a,b = nombres multiprcision, retourne a + b *)
let prefix ++ a b = add_c a b 0;;

                       (* +----------------+
                          |  Soustraction  |
                          +----------------+ *)

(* a,b = nombres multiprcision, c = retenue, retourne a - b + c *)
let rec sub_c a b c =
  if (a = zro) & (b = zro) then nombre_of_int(c)
  else begin
    let a0,a1 = dcompose(a) in
    let b0,b1 = dcompose(b) in
    let x     = a0 - b0 + c  in
    if x >= 0 then assemble x          (sub_c a1 b1 0)
              else assemble (x + BASE) (sub_c a1 b1 (-1))
  end
;;


(* a,b = nombres multiprcision, retourne a - b *)
let prefix -- a b = sub_c a b 0;;


                        (* +---------------+
                           |  Comparaison  |
                           +---------------+ *)

(* a,b = nombres multiprcision, compare a  b *)

let rec cmp a b = match a.lg,b.lg with
  |0,0 -> Egal
  |_,0 -> Suprieur
  |0,_ -> Infrieur
  |_    -> let a0,a1 = dcompose(a) in
           let b0,b1 = dcompose(b) in
           match cmp a1 b1 with
           |Egal ->  if a0 < b0 then Infrieur else if a0 > b0 then Suprieur else Egal
           |r -> r
;;

                      (* +------------------+
                         |  Multiplication  |
                         +------------------+ *)

(* a = nombre multiprcision, b = chiffre c = retenue, retourne a*b + c *)
let rec mul_1 a b c =
  if (a = zro) or (b = 0) then nombre_of_int(c)
  else begin
    let a0,a1 = dcompose(a) in
    let x     = a0*b + c     in
    assemble (x mod BASE) (mul_1 a1 b (x/BASE))
  end
;;

(* a,b = nombres multiprcision, retourne a*b (algorithme naf) *)
let rec mul a b =
  if b = zro then zro
  else let b0,b1 = dcompose(b) in (mul_1 a b0 0) ++ (assemble 0 (mul a b1))
;;

(* a,b = nombres multiprcision, retourne a*b (algorithme de Knuth) *)
let seuil = 30;;
let rec mulk a b =

  if b.lg < seuil then mul a b

  else begin
    let n     = b.lg/2 in
    let a0,a1 = a << n
    and b0,b1 = b << n in
    let x     = mulk a0 b0
    and y     = mulk a1 b1  in
    let z     = (mulk (a0 ++ a1) (b0 ++ b1)) -- (x ++ y) in
    x ++ ((z ++ (y >> n)) >> n)
  end
;;

(* essai et chronomtrage *)
let test(n) =
  let a = random_nombre(n) and b = random_nombre(n) in

  let t = sys__time()      in
  let c = mul a b          in
  let t1= sys__time() -. t in

  let t = sys__time()      in
  let d = mulk a b         in
  let t2= sys__time() -. t in

  (cmp c d), t1, t2
;;

test 1000;; (* Egal, 18.72, 4.04 *)

let prefix ** = mulk;;


                         (* +------------+
                            |  Division  |
                            +------------+ *)


(* a = nombre multiprcision, b entier compris entre 1 et BASE^2-1 *)
(* retourne (a mod b), (a/b)                                           *)
let rec div_1 a b =
  if a = zro then 0,zro
  else begin
    let a0,a1 = dcompose(a) in
    let r,q   = div_1 a1 b   in
    let x     = a0 + BASE*r  in
    (x mod b), assemble (x/b) q
  end
;;

(* a,b = nombres multiprcision, b > 0, retourne (amod b), (a/b) *)
let rec prefix // a b =
  let n = (b.lg - 1)/2 in

  if n = 0
  then let r,q = div_1 a (int_of_nombre b) in nombre_of_int(r),q

  else let a0,a1 = a << n in
  match cmp a1 b with

  |Suprieur
  |Egal     -> let r1,q1 = a1 // b in
               let r0,q0 = (a0 ++ (r1 >> n)) // b in
               r0, (q0 ++ (q1 >> n))

  |Infrieur ->let b0,b1 = b << n   in
               let r,q   = a1 // b1 in
               let x     = a0 ++ (r >> n)
               and y     = b0 ** q in
               if cmp x y = Infrieur then (x ++ b -- y), (q -- un)
                                      else (x -- y), q
    
;;

(* a,b = nombres multiprcision, n = taille *)
(* retourne floor(BASE^n*(a/b))             *)
let floor a b n = let _,q = (a >> n) // b in q;;

(* a,b = nombres multiprcision, n = taille *)
(* retourne ceil(BASE^n*(a/b))              *)
let ceil a b n = let r,q = (a >> n) // b in if r = zro then q else q ++ un;;

floor un (nombre_of_int 49) 20;;
ceil  un (nombre_of_int 49) 20;;



                     (* +---------------------+
                        |  Calcul de sqrt(2)  |
                        +---------------------+ *)


(* n = nombre de chiffres voulus *)
let rac2 n =

  let rec loop a b c d =
    if 2*c.lg > n then a,b,c,d
    else let x = b**c and y = a++d in loop (a**a ++ x) (b**y) (c**y) (x++d**d)
  in

  let a,b,c,d = loop zro un un deux in
  (floor (b++d) d n), (ceil (a++c) c n)
;;
 
let u,v = rac2 100 in
  affiche(u);
  affiche(v);
  affiche(u**u);
  affiche(v**v)
;;


                  (* +--------------------------+
                     |  Racine carre gnrale  |
                     +--------------------------+ *)


(* a = entier compris entre 0 et BASE^4 - 1 *)
(* retourne floor(sqrt(a))                  *)
let sqrt_1 a =
  if a = 0 then 0
  else let rec loop(x) = let y = (x + a/x)/2 in
                         if y >= x then x else loop(y)
  in loop(a)
;;

(* a = nombre multiprcision                 *)
(* retourne b = floor(sqrt a) et r = a - b^2 *)
let rec sqrt a =
  let n = (a.lg - 1)/4 in

  if n = 0 then let b = nombre_of_int(sqrt_1 (int_of_nombre a)) in (a -- b**b), b

  else begin
    let a0,a1 = a  << n in
    let a1,a2 = a1 << n in
    let r1,b1 = sqrt a2 in
    let r0,b0 = (a1 ++ (r1 >> n)) // (b1 ++ b1) in
    let u     = a0 ++ (r0 >> n)
    and v     = b0 ** b0
    and b     = b0 ++ (b1 >> n) in
    match cmp u v with
    |Suprieur
    |Egal       -> (u -- v), b
    |Infrieur  -> (u ++ b ++ b -- un -- v), (b -- un)
  end
;;
       
let r,b = sqrt(deux >> 2000) in
  affiche(b);
  affiche(b**b ++ r);
  affiche((b++un)**(b++un))
;;


                  (* +--------------------------+
                     |  Comparaison de vitesse  |
                     +--------------------------+ *)


(* fraction continue, avec une seule division *)
let r2 n =

  let rec loop a b c d =
    if 4*c.lg > n then a,b,c,d
    else let x = b**c and y = a++d in loop (a**a ++ x) (b**y) (c**y) (x++d**d)
  in

  let a,b,c,d = loop zro un un deux in
  let b' = b**(a++d) and d' = b**c ++ d**d in
  floor (b'++d') d' n
;;

let test(n) =

  let t    = sys__time()          in
  let r1   = r2(n)                in
  let t1   = sys__time() -. t     in

  let t    = sys__time()          in
  let _,r2 = sqrt (deux >> (2*n)) in
  let t2   = sys__time() -. t     in

  (cmp r1 r2), t1, t2
;;

test 100;; (* Egal,  0.4,   0.17 *)
test 200;; (* Egal,  1.21,  0.5  *)
test 400;; (* Egal,  3.7,   1.5  *)
test 800;; (* Egal, 11.26,  4.53 *)
